#!/usr/bin/env python

# This script is hereby placed in the Public Domain
#
# Convert Raw FDR State Transition output to a format that 'dot' can recognise
# takes an input file containing the raw data as its only argument
# input file should be of the form:
#
# 1.  PROCESS_NAME
# 2.  INITIAL_STATE
# 3.  TRANSITIONS
# 4.  EVENT_NUMBERS
# 5.  EVENT_NAMES
# 6.  ACCEPTANCES
# 7.  DIVERGENCES

# where PROCESS_NAME is the name of the process that FDR has produced the
# transition diagram for, INITIAL_STATE is its initial state as an integer,
# TRANSITION is a line of 3-typles separated by single spaces corresponding
# to the FDR output of the 'transitions' command. Each 3-tuple is a
# transition for the state machine of the process and looks like:
# "{from event_no to}" where 'from' and 'to' are states within the state
# machine and 'event_no' is an event number that when performed in state
# 'from' transitions the machine to the state 'to'.
# EVENT_NUMBERS is a line of numbers separated by spaces
# EVENT_NAMES is a line of event names separated by spaces. The ith event
# number corresponds to the ith event name.
# ACCEPTANCES is a line of acceptance sets. The ith set corresponds to
# the acceptances of state i (indexed from 0)
# DIVERGENCES is a line of booleans (0 or 1). The ith boolean indicates whether
# node i is divergent.
#
# see the GraphViz documentation for the dot input format.
#
# Warning: This code is an awful hack and ought to be cleaned up properly.
#          My apologies in advance.

import sys
import getopt

if sys.version_info[0] == 2 and sys.version_info[1] < 3:
	True = 1
	False = 0


errorInArgs = False
labelDivergences = False
labelAcceptances = False
argc = len(sys.argv)
if argc >= 2:
	if argc > 2:
		options = sys.argv[2:]
		(opts, rest) = getopt.getopt(options, "da")
		if rest == []:
			for (option, value) in opts:
				if (option == "-d"):
					labelDivergences = True

				elif option == "-a":
					labelAcceptances = True

				else:
					errorInArgs = True
		else:
			errorInArgs = True
else:
	errorInArgs = True

if errorInArgs:
	sys.stderr.write("Usage: %s inputfile [-da]\n" % sys.argv[0])
	sys.stderr.write("   -d: label divergent nodes\n")
	sys.stderr.write("   -a: label each node with its acceptances\n")
	sys.exit(1)
	
if labelDivergences:
	sys.stderr.write("(Divergent nodes will be labelled as such.)\n")
if labelAcceptances:
	sys.stderr.write("(Nodes will be labelled with acceptances.)\n")

inputfile = sys.argv[1]

fh = open(inputfile)
processName = fh.readline()
initialState = fh.readline()
transitions = fh.readline()
eventNumbers = fh.readline()
eventNames = fh.readline()
acceptancesLine = fh.readline()
divergencesLine = fh.readline()

# remove any trailing junk from processName and initialState
processName = processName.split()[0]
initialState = initialState.split()[0]

eventNumbers = eventNumbers.split()
eventNames = eventNames.split()

# 'events' maps event numbers to event names
# add default events
events = {"0": "_tau", "1": "_tick"}

# populate the 'events' mapping
for index in range(len(eventNumbers)):
	events[eventNumbers[index]] = eventNames[index]	

acceptances = {}
divergences = {}

# parse the divergences
divergencesLine = divergencesLine.split(" ")
numStates = len(divergencesLine)
state = 0
for div in divergencesLine:
	div = div.strip()
	if div == "0":
		divergences[state] = False
	else:
		assert div == "1"
		divergences[state] = True
	state += 1
assert numStates == state

# parse the acceptances
def nextchar(str):
	return (str[0],str[1:])

def peekchar(str):
	if len(str) > 0:
		return str[0]
	else:
		return ""


# a grammar for acceptances -- we hve one <acceptancesSetSet> per node
# <acceptancesLine>   ::= <acceptancesSetSet>(' '<acceptancesSetSet>)*'\n'
# <acceptancesSetSet> ::= '{'[<acceptancesSet>(' '<acceptancesSet>)*]'}' | 
#                         <singleton>
# <acceptancesSet>    ::= '{'[<element>(' '<element>)*]'}' | <singleton>
# <singleton>         ::= <element>
# <element>           ::= ([1-9])*[0-9]
#
# notice that the grammar for <acceptancesSetSet> and <acceptancesSet> are
# identical accept that <acceptancesSet> is replaced by <element>.
# hence we write one parameterised function to handle both

# populates the 'acceptances' structure. returns the number of
# <acceptancesSetSet>s parsed
def parseAcsLine(acsLine):
	acsCount = 0

	# remove trailing newline etc. 
	acsLine = acsLine.strip()
	
	(ac,acsLine) = parseAcsSetSet(acsLine)
	acceptances[acsCount] = ac
	acsCount+=1
	while acsLine != "":
		(c,acsLine) = nextchar(acsLine)
		assert c == ' '
		(ac,acsLine) = parseAcsSetSet(acsLine)
		acceptances[acsCount] = ac
		acsCount += 1
	return acsCount

# a general function to parse a <set> that adheres to the following gramamr
# <set>               ::= '{'[<thing>(' '<thing>)*]'}' | <singleton>
# <singleton>         ::= <element>
#
# 'parseThing' knows how to parse <thing>s
# 'transformSingleton' takes a parsed <singleton> and transforms it
#                      so that it can be interpreted as a <set>
def parseSet(parseThing, transformSingleton, acsLine):
	res = []
	c = peekchar(acsLine)
	if c == '{':
		(c,acsLine) = nextchar(acsLine)
		assert c == '{'
		if peekchar(acsLine) == '}':
			(c,acsLine) = nextchar(acsLine)
			assert c == '}'
			return (res,acsLine)
		else:
			(thing,acsLine) = parseThing(acsLine)
			res.append(thing)
			(c,acsLine) = nextchar(acsLine)
			while c == ' ':
				(thing,acsLine) = parseThing(acsLine)
				res.append(thing)
				(c,acsLine) = nextchar(acsLine)
			assert c == '}'
			return (res,acsLine)
	else:
		(singleton,acsLine) = parseElement(acsLine)
		res.append(transformSingleton(singleton))
		return (res,acsLine)

# returns a pair (ac,acsLine) where 'ac' is the <acceptancesSetSet> that
# was parsed and acsLine is the remaining text
# - 'ac' is a list of lists of events
def parseAcsSetSet(acsLine):
	(setset,line) = parseSet(parseAcsSet,lambda s : [s],acsLine)
	return (setset,line)

# return a pair (set,acsLine) where 'set' is <acceptancesSet> in the form
# of a list of events
def parseAcsSet(acsLine):
	(set,line) = parseSet(parseElement,lambda s : s,acsLine)
	return (set,line)
	
# parses an element, return (element,acsLine)
def parseElement(acsLine):
	element = ""
	(c,acsLine) = nextchar(acsLine)
	assert c in '0123456789'
	element += c
	pk = peekchar(acsLine)
	while pk in '0123456789' and pk != '':
		(c,acsLine) = nextchar(acsLine)
		assert c == pk
		element += c
		pk = peekchar(acsLine)
	return (events[element],acsLine)
	
		

acsSetSetsParsed = parseAcsLine(acceptancesLine)
assert acsSetSetsParsed == numStates

def acceptancesStr(accs):
	res = ""
	for acc in accs:
		res += "{"
		index = 0
		while index != len(acc):
			res += acc[index]
			index += 1
			if index < len(acc):
				res += ","
		res += "}"
	return res

# produce the dot output
print "digraph \"" + processName + "\" {"

state = 0
while state != numStates:
	label = ""
	if labelAcceptances:
		label += acceptancesStr(acceptances[state])
	if labelDivergences and divergences[state]:
		label += "\\n(diverges)"
	color = "auto"
	if state == int(initialState):
		color = "red"
		
	print "node [color=" + color +", label=\"" + label + "\"]; " + str(state) + ";"
	state += 1

for t in transitions.split('{'):
	event = t.split()
        if len(event) == 3:
                print event[0] + " -> " \
                + event[2].split('}')[0] + " [label=\"" + events[event[1]] + "\"];"

print "}"
